from typing import TYPE_CHECKING, Annotated, Union

import strawberry
from strawberry.federation.schema_directives import Key
from strawberry.unset import UNSET

from oauth2_lib.strawberry import authenticated_field
from orchestrator.domain.base import ProductModel
from orchestrator.graphql.pagination import Connection
from orchestrator.graphql.schemas.fixed_input import FixedInput
from orchestrator.graphql.schemas.product_block import ProductBlock
from orchestrator.graphql.schemas.workflow import Workflow
from orchestrator.graphql.types import GraphqlFilter, GraphqlSort, OrchestratorInfo
from orchestrator.schemas.product import ProductSchema

if TYPE_CHECKING:
    from orchestrator.graphql.schemas.subscription import SubscriptionInterface


federation_key_directives = [Key(fields="productId", resolvable=UNSET)]


@strawberry.experimental.pydantic.type(model=ProductSchema, directives=federation_key_directives)
class ProductType:
    product_id: strawberry.auto
    name: strawberry.auto
    description: strawberry.auto
    product_type: strawberry.auto
    status: strawberry.auto
    tag: strawberry.auto
    created_at: strawberry.auto
    end_date: strawberry.auto
    product_blocks: list[ProductBlock]
    fixed_inputs: list[FixedInput]
    workflows: list[Workflow]

    @authenticated_field(description="Returns list of subscriptions of the product type")  # type: ignore
    async def subscriptions(
        self,
        info: OrchestratorInfo,
        filter_by: Union[list[GraphqlFilter], None] = None,
        sort_by: Union[list[GraphqlSort], None] = None,
        first: int = 10,
        after: int = 0,
    ) -> Connection[Annotated["SubscriptionInterface", strawberry.lazy(".subscription")]]:
        from orchestrator.graphql.resolvers.subscription import resolve_subscriptions

        filter_by_with_related_subscriptions = (filter_by or []) + [GraphqlFilter(field="product", value=self.name)]
        return await resolve_subscriptions(info, filter_by_with_related_subscriptions, sort_by, first, after)


@strawberry.experimental.pydantic.type(model=ProductModel, all_fields=True)
class ProductModelGraphql:
    pass



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from orchestrator.cli.generator.generator.fixed_input import (
    get_int_enum_fixed_inputs,
    get_non_standard_fixed_inputs,
    get_str_enum_fixed_inputs,
    replace_enum_fixed_inputs,
)
from orchestrator.cli.generator.generator.helpers import get_product_file_name, path_to_module, product_types_module
from orchestrator.cli.generator.generator.settings import product_generator_settings


def get_product_path(config: dict) -> str:
    file_name = get_product_file_name(config)
    return f"{product_generator_settings.PRODUCT_TYPES_PATH}/{file_name}.py"


def generate_product(context: dict) -> None:
    config = context["config"]
    environment = context["environment"]
    writer = context["writer"]

    product = config["type"]
    fixed_inputs = config.get("fixed_inputs", [])
    product_blocks = config.get("product_blocks", [])

    non_standard_fixed_inputs = get_non_standard_fixed_inputs(fixed_inputs)
    int_enums = get_int_enum_fixed_inputs(fixed_inputs)
    str_enums = get_str_enum_fixed_inputs(fixed_inputs)

    template = environment.get_template("product.j2")
    content = template.render(
        product=product,
        product_blocks_module=path_to_module(product_generator_settings.PRODUCT_BLOCKS_PATH),
        product_types_module=product_types_module,
        non_standard_fixed_inputs=non_standard_fixed_inputs,
        fixed_inputs=replace_enum_fixed_inputs(fixed_inputs),
        product_blocks=product_blocks,
        int_enums=int_enums,
        str_enums=str_enums,
    )

    path = get_product_path(config)
    writer(path, content)



import json

from deepdiff import DeepDiff


def assert_no_diff(expected, actual, exclude_paths=None):
    diff = DeepDiff(expected, actual, ignore_order=True, exclude_paths=exclude_paths)
    prettydiff = f"Difference: {json.dumps(diff, indent=2, default=lambda x: str(x))}"

    assert diff == {}, f"Difference between expected and actual output\n{prettydiff}"



from typing import Callable

import structlog

from orchestrator.db import ProductTable, WorkflowTable
from orchestrator.db.database import SearchQuery
from orchestrator.db.filters import generic_filter
from orchestrator.db.filters.generic_filters import generic_is_like_filter, generic_range_filters

logger = structlog.get_logger(__name__)

start_date_range_filters = generic_range_filters(WorkflowTable.created_at)


def products_filter(query: SearchQuery, value: str) -> SearchQuery:
    """Filter ProductBlocks by '-'-separated list of Product 'name' (column) values."""
    products = value.split("-")
    return query.filter(WorkflowTable.products.any(ProductTable.name.in_(products)))


WORKFLOW_FILTER_FUNCTIONS_BY_COLUMN: dict[str, Callable[[SearchQuery, str], SearchQuery]] = {
    "workflowId": generic_is_like_filter(WorkflowTable.workflow_id),
    "name": generic_is_like_filter(WorkflowTable.name),
    "description": generic_is_like_filter(WorkflowTable.description),
    "createdAt": generic_is_like_filter(WorkflowTable.created_at),
    "products": products_filter,
} | start_date_range_filters

filter_workflows = generic_filter(WORKFLOW_FILTER_FUNCTIONS_BY_COLUMN)



from sqlalchemy.inspection import inspect

from orchestrator.db import ProcessTable
from orchestrator.db.sorting.product import PRODUCT_SORT_FUNCTIONS_BY_COLUMN
from orchestrator.db.sorting.sorting import generic_column_sort, generic_sort
from orchestrator.utils.helpers import to_camel

PROCESS_CAMEL_SORT = {to_camel(key): generic_column_sort(value) for key, value in inspect(ProcessTable).columns.items()}
PROCESS_SNAKE_SORT = {key: generic_column_sort(value) for key, value in inspect(ProcessTable).columns.items()}
PROCESS_PRODUCT_SORT = {
    to_camel(key if "product" in key else f"product_{key}"): value
    for key, value in PRODUCT_SORT_FUNCTIONS_BY_COLUMN.items()
}

PROCESS_SORT_FUNCTIONS_BY_COLUMN = PROCESS_PRODUCT_SORT | PROCESS_SNAKE_SORT | PROCESS_CAMEL_SORT
PROCESS_SORT_FUNCTIONS_BY_COLUMN["workflow"] = generic_column_sort(
    ProcessTable.workflow_name
)  # TODO: deprecated, remove in 1.3
PROCESS_SORT_FUNCTIONS_BY_COLUMN["status"] = generic_column_sort(
    ProcessTable.last_status
)  # TODO: deprecated, remove in 1.3
PROCESS_SORT_FUNCTIONS_BY_COLUMN["creator"] = generic_column_sort(
    ProcessTable.created_by
)  # TODO: deprecated, remove in 1.3
PROCESS_SORT_FUNCTIONS_BY_COLUMN["started"] = generic_column_sort(
    ProcessTable.started_at
)  # TODO: deprecated, remove in 1.3
PROCESS_SORT_FUNCTIONS_BY_COLUMN["modified"] = generic_column_sort(
    ProcessTable.last_modified_at
)  # TODO: deprecated, remove in 1.3

sort_processes = generic_sort(PROCESS_SORT_FUNCTIONS_BY_COLUMN)



from sqlalchemy.inspection import inspect

from orchestrator.db import ProductTable
from orchestrator.db.sorting.sorting import generic_column_sort, generic_sort
from orchestrator.utils.helpers import to_camel

PRODUCT_SORT_FUNCTIONS_BY_COLUMN = {
    to_camel(key): generic_column_sort(value) for [key, value] in inspect(ProductTable).columns.items()
}

sort_products = generic_sort(PRODUCT_SORT_FUNCTIONS_BY_COLUMN)
